#
"""template - Template utilities for hubugs."""
# Copyright © 2011-2018  James Rowe <jnrowe@gmail.com>
#
# SPDX-License-Identifier: GPL-3.0+
#
# This file is part of hubugs.
#
# hubugs is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# hubugs is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# hubugs.  If not, see <http://www.gnu.org/licenses/>.

import datetime
import operator
import os
import sys

from typing import Callable, Dict, List, Optional

import click
import html2text as html2
import jinja2
import misaka

from jnrbase import xdg_basedir
from jnrbase.colourise import success
from jnrbase.human_time import human_timestamp
from pygments import highlight as pyg_highlight
from pygments.formatters import get_formatter_by_name
from pygments.lexers import get_lexer_by_name

from . import utils


PKG_DATA_DIRS = [os.path.join(xdg_basedir.user_data('hubugs'), 'templates'), ]
for directory in xdg_basedir.get_data_dirs('hubugs'):
    PKG_DATA_DIRS.append(os.path.join(directory, 'templates'))

ENV = jinja2.Environment(loader=jinja2.ChoiceLoader(
    [jinja2.FileSystemLoader(s) for s in PKG_DATA_DIRS]))
ENV.loader.loaders.append(jinja2.PackageLoader('hubugs', 'templates'))
ENV.filters['relative_time'] = human_timestamp


class EmptyMessageError(ValueError):

    """Error to raise when the user provides an empty message."""

    pass


def get_template(__group: str, __name: str) -> jinja2.environment.Template:
    """Fetch a Jinja template instance.

    Args:
        __group: Template group identifier
        __name: Template name

    Returns:
        Jinja template instance
    """
    template_set = utils.get_git_config_val('hubugs.templates', 'default')
    return ENV.get_template('/'.join([template_set, __group, __name]))


def jinja_filter(__func: Callable) -> Callable:
    """Simple decorator to add a new filter to Jinja environment.

    Args:
        __func: Function to add to Jinja environment

    Returns:
        Unmodified function
    """
    ENV.filters[__func.__name__] = __func

    return __func


@jinja_filter
def colourise(__text: str, fg: Optional[str] = None, bg: Optional[str] = None,
              **kwargs)-> str:
    """Colourise text.

    Returns text untouched if colour output is not enabled

    Args:
        __text: Text to colourise
        fg: Foreground colour
        bg: Background colour
        kwargs: Formatting to apply to text

    Returns:
        Colourised text, when possible
    """
    return click.style(__text, fg, bg, **kwargs)
# American spelling, just for Brandon Cady ;)
ENV.filters['colorize'] = ENV.filters['colourise']


@jinja_filter
def highlight(__text: str, lexer: Optional[str] = 'diff',
              formatter: Optional[str] = 'terminal') -> str:
    """Highlight text with pygments.

    Returns text untouched if colour output is not enabled

    Args:
        __text: Text to highlight
        lexer: Jinja lexer to use
        formatter: Jinja formatter to use

    Returns:
        Syntax highlighted output, when possible
    """
    if sys.stdout.isatty():
        lexer = get_lexer_by_name(lexer)
        formatter = get_formatter_by_name(formatter)
        return pyg_highlight(__text, lexer, formatter)
    else:
        return __text


@jinja_filter
def html2text(__html: str, width: Optional[int] = 80,
              ascii_replacements: Optional[bool] = False) -> str:
    """HTML to plain text renderer.

    Args:
        __html: Text to process
        width: Paragraph width
        ascii_replacements: Use psuedo-ascii replacements for Unicode

    Returns:
        Rendered text
    """
    html2.BODY_WIDTH = width
    html2.UNICODE_SNOB = ascii_replacements
    return html2.html2text(__html).strip()


@jinja_filter
def markdown(__text: str) -> str:
    """Markdown to HTML renderer.

    Args:
        __text: Text to process

    Returns:
        Rendered HTML
    """
    extensions = misaka.EXT_AUTOLINK | misaka.EXT_FENCED_CODE
    return misaka.html(__text, extensions, misaka.HTML_SKIP_HTML)


def display_bugs(__bugs: List[Dict[str, str]], __order: str, **extras) -> str:
    """Display bugs to users.

    Args:
        __bugs: Bugs to display
        __order: Sorting order for displaying bugs
        extras: Additional values to pass to templates

    Returns:
        Rendered template output
    """
    if not __bugs:
        return success('No bugs found!')

    # Match ordering method to bug attribute
    if __order == 'updated':
        attr = 'updated_at'
    else:
        attr = __order

    __bugs = sorted(__bugs, key=operator.attrgetter(attr))

    # Default to 80 columns, when stdout is not a tty
    columns = click.get_terminal_size()[0]

    template = get_template('view', 'list.txt')

    max_id = max(i.number for i in __bugs)
    id_len = len(str(max_id))
    spacer = ' ' * (id_len - 2)

    return template.render(bugs=__bugs, spacer=spacer, id_len=id_len,
                           max_title=columns - id_len - 2, **extras)


def edit_text(edit_type: Optional[str] = 'default',
              data: Optional[str] = None) -> str:
    """Edit data with external editor.

    Args:
        edit_type: Template to use in editor
        data: Information to pass to template

    Returns:
        User supplied text

    Raises:
        EmptyMessageError: No message given
        EmptyMessageError: Message not edited
    """
    template = get_template('edit', '{}.mkd'.format(edit_type))
    comment_char = utils.get_git_config_val('core.commentchar', '#')
    if not data:
        data = {}
    data['comment_char'] = comment_char

    text = click.edit(template.render(data), require_save=True,
                      extension='.mkd')
    if text:
        text = ''.join(filter(lambda s: not s.startswith(comment_char),
                              text.splitlines())).strip()

    if not text:
        raise EmptyMessageError('No message given')

    return text.strip()



#
"""test_utils - Test utility functions."""
# Copyright © 2011-2018  James Rowe <jnrowe@gmail.com>
#
# SPDX-License-Identifier: GPL-3.0+
#
# This file is part of hubugs.
#
# hubugs is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# hubugs is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# hubugs.  If not, see <http://www.gnu.org/licenses/>.

from subprocess import CalledProcessError
from typing import Optional

from click import BadParameter
from pytest import mark, raises

from hubugs import ProjectNameParamType
from hubugs import utils


@mark.parametrize('repo, expected', [
    ('misc-overlay', 'JNRowe/misc-overlay'),
    ('JNRowe/misc-overlay', 'JNRowe/misc-overlay'),
    ('ask/python-github2', 'ask/python-github2'),
])
def test_ProjectNameParamType_repo_name(repo: str, expected: str, monkeypatch):
    monkeypatch.setattr(
        'hubugs.utils.get_github_api',
        lambda: utils.AttrDict(repos=utils.AttrDict(show=lambda: True)))
    monkeypatch.setattr('hubugs.utils.get_git_config_val', lambda _: 'JNRowe')

    p = ProjectNameParamType()
    p.convert(repo, None, None) == expected


def test_ProjectNameParamType_no_user(monkeypatch):
    monkeypatch.setattr(
        'hubugs.utils.get_github_api',
        lambda: utils.AttrDict(repos=utils.AttrDict(show=lambda: True)))
    monkeypatch.setattr('hubugs.utils.get_git_config_val', lambda _: None)

    p = ProjectNameParamType()
    with raises(BadParameter):
        p.convert('misc-overlay', None, None)


def test_GetGitConfigVal_valid_key(monkeypatch):
    monkeypatch.setattr('subprocess.check_output',
                        lambda *args, **kwargs: b'JNRowe')

    assert utils.get_git_config_val('github.user') == 'JNRowe'


def test_GetGitConfigVal_invalid_key(monkeypatch):
    monkeypatch.setattr('subprocess.check_output', lambda *args, **kwargs: b'')

    assert utils.get_git_config_val('no_such_key') == ''


def test_GetGitConfigVal_command_error(monkeypatch):
    def raise_error(*args, **kwargs):
        raise CalledProcessError('255', 'cmd')
    monkeypatch.setattr('subprocess.check_output', raise_error)

    assert utils.get_git_config_val('github.user') is None


def test_GetEditor_git_editor_envvar(monkeypatch):
    monkeypatch.setenv('EDITOR', 'custom git editor')
    assert utils.get_editor() == ['custom', 'git', 'editor']


def test_GetEditor_editor_environment_editor(monkeypatch):
    monkeypatch.setenv('EDITOR', 'editor')
    assert utils.get_editor() == ['editor', ]


@mark.parametrize('repo', [
    'git@github.com:JNRowe/misc-overlay.git',
    'git@github.com:JNRowe/misc-overlay',
    'git://github.com/JNRowe/misc-overlay.git',
    'git://github.com/JNRowe/misc-overlay',
    'https://JNRowe@github.com/JNRowe/misc-overlay.git',
    'https://JNRowe@github.com/JNRowe/misc-overlay',
    'http://JNRowe@github.com/JNRowe/misc-overlay.git',
    'http://JNRowe@github.com/JNRowe/misc-overlay',
    'http://github.com/JNRowe/misc-overlay.git',
    'http://github.com/JNRowe/misc-overlay',

    # hg-git
    'git+ssh://git@github.com:JNRowe/misc-overlay.git',
])
def test_GetRepo_repo_url(repo: str, monkeypatch):
    # side_effect to skip hubugs.project call
    results = [repo, None]
    monkeypatch.setattr('hubugs.utils.get_git_config_val',
                        lambda *args, **kwargs: results.pop())

    assert utils.get_repo() == 'JNRowe/misc-overlay'


@mark.parametrize('repo', [
    'git://github.com/misc-overlay.git',
    None,
    'http://example.com/dog.git',
])
def test_GetRepo_broken_url(repo: Optional[str], monkeypatch):
    # side_effect to skip hubugs.project call
    results = [repo, None]
    monkeypatch.setattr('hubugs.utils.get_git_config_val',
                        lambda *args, **kwargs: results.pop())

    with raises(ValueError):
        utils.get_repo()


def test_GetRepo_config_project(monkeypatch):
    monkeypatch.setattr('hubugs.utils.get_git_config_val',
                        lambda *args, **kwargs: 'JNRowe/misc-overlay')

    assert utils.get_repo() == 'JNRowe/misc-overlay'



#
"""models - GitHub models for hubugs."""
# Copyright © 2012-2018  James Rowe <jnrowe@gmail.com>
#
# SPDX-License-Identifier: GPL-3.0+
#
# This file is part of hubugs.
#
# hubugs is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# hubugs is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# hubugs.  If not, see <http://www.gnu.org/licenses/>.

import collections
import contextlib
import datetime

from typing import Dict, Optional

from jnrbase.iso_8601 import parse_datetime

# We used to use tight, explicit bindings for API objects but the desire to
# support Python 3 and the lack of a usable binding library has made it
# necessary to use the loose dynamic binding implemented below.
#
# I fully expect to return to strict binding in the future, once either a tool
# becomes available or I free up a little more itch-scratching time


def object_hook(__d: Dict[str, str], __name: Optional[str] = 'unknown'):
    """JSON object hook to create dot-accessible objects.

    Args:
        __d: Dictionary to operate on
        name: Fallback name, if dict has no ``type`` key
    """
    # FIXME: Dump _links attributes for the time being
    if '_links' in __d:
        __d.pop('_links')
    for k, v in __d.items():
        with contextlib.suppress(TypeError, ValueError):
            __d[k] = parse_datetime(v).replace(tzinfo=None)
    return collections.namedtuple(__d.get('type', __name), __d.keys())(**__d)


def _v2_conv_timestamp(__s: str):
    """Parse API v2 style timestamps.

    Args:
        __s: Timestamp to parse
    """
    stamp = parse_datetime(__s).astimezone(datetime.timezone.utc)
    return stamp.isoformat()[:-6] + 'Z'


def from_search(__obj):
    """Support legacy API search results as API v3 issues(-ish).

    This is an awful hack to workaround the lack of search support in API
    v3.  It needs to be removed at the first possible opportunity.

    Returns:
        Issue: API v2 issue mangled to look like a API v3 result
    """

    avatar_url = ('https://secure.gravatar.com/avatar/{}?d='
                  'https://a248.e.akamai.net/assets.github.com%2F'
                  'images%2Fgravatars%2Fgravatar-140.png')
    label_url = 'https://api.github.com/repos/{}/{}/labels/{}'

    owner, project = __obj.html_url.split('/')[3:5]

    d = __obj.__dict__
    d['id'] = '<from search>'
    d['user'] = {
        'avatar_url': avatar_url.format(__obj.gravatar_id),
        'gravatar_id': __obj.gravatar_id,
        'login': __obj.user,
        'url': 'https://api.github.com/users/{}'.format(__obj.user),
    }
    if 'closed_by' in __obj._fields:
        d['closed_by'] = {
            'avatar_url': avatar_url.format(__obj.closed_by),
            'gravatar_id': __obj.gravatar_id,
            'login': __obj.user,
            'url': 'https://api.github.com/users/{}'.format(__obj.user),
        }
        d['closed_at'] = _v2_conv_timestamp(__obj.closed_at)
    d['created_at'] = _v2_conv_timestamp(__obj.created_at)
    d['updated_at'] = _v2_conv_timestamp(__obj.updated_at)
    if 'labels' in __obj._fields:
        labels = []
        for name in __obj.labels:
            labels.append({
                'color': '000000',
                'name': name,
                'url': label_url.format(owner, project, name)
            })
        d['labels'] = labels
    return object_hook(d, 'issue')



#
"""hubugs - Simple client for GitHub issues.

``hubugs`` is a very simple client for working with `GitHub’s issue tracker`_.

.. _GitHub’s issue tracker: http://github.com/blog/411-github-issue-tracker
"""
# Copyright © 2010-2018  James Rowe <jnrowe@gmail.com>
#                        Ben Griffiths
#                        Matt Leighton <mleighy@gmail.com>
#
# SPDX-License-Identifier: GPL-3.0+
#
# This file is part of hubugs.
#
# hubugs is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# hubugs is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# hubugs.  If not, see <http://www.gnu.org/licenses/>.

from . import _version


__version__ = _version.dotted
__date__ = _version.date
__copyright__ = '2010-2016  James Rowe'


# This is here to workaround UserWarning messages caused by path fiddling in
# dependencies
try:
    import pkg_resources  # NOQA: F401
except ImportError:
    pass

import atexit
import errno
import getpass
import logging
import os
# Used by raw_input, when imported
import readline  # NOQA: F401
import sys

from base64 import b64encode
from typing import Callable, List, Optional

import click
import httplib2

from jnrbase.attrdict import AttrDict
from jnrbase.colourise import fail, success, warn


logging.basicConfig(level=logging.ERROR, format='%(asctime)s - %(message)s',
                    datefmt='%FT%T')
atexit.register(logging.shutdown)


from . import (template, utils)


class ProjectNameParamType(click.ParamType):

    """Project name parameter parser."""

    name = 'project'

    def convert(self, __value: str, __param: Optional[click.Argument],
                __ctx: Optional[click.Context]) -> str:
        """Set fully qualified GitHub project name."""
        if '/' not in __value:
            user = os.getenv('GITHUB_USER',
                             utils.get_git_config_val('github.user'))
            if not user:
                raise click.BadParameter('No GitHub user setting!')
            __value = '/'.join([user, __value])
        return __value


@click.group(help='Simple client for GitHub issues.',
             epilog='Please report bugs at '
                    'https://github.com/JNRowe/hubugs/issues',
             context_settings={'help_option_names': ['-h', '--help']})
@click.version_option(_version.dotted)
@click.option('--pager/--no-pager', help='Pass output through a pager.')
@click.option('-p', '--project', type=ProjectNameParamType(),
              default=utils.get_repo(),
              help='GitHub project to operate on.')
@click.option('-u', '--host-url',
              default=utils.get_git_config_val('hubugs.host-url',
                                               'https://api.github.com'),
              help='GitHub Enterprise host to connect to.')
@click.pass_context
def cli(ctx: click.Context, pager: bool, project: str, host_url: str):
    """Main command entry point.

    Args:
        ctx: Current command context
        pager: Whether to page output
        project: GitHub project name
        host: Hostname to connect to
    """
    ctx.obj = utils.setup_environment(project, host_url)
    ctx.obj.update({
        'host_url': host_url,
        'pager': pager,
        'project': project,
    })


# Convenience wrappers for defining command arguments
def bugs_parser(__f: Callable) -> Callable:
    __f = click.argument('bugs', nargs=-1, type=click.INT)(__f)
    return __f


def message_parser(__f: Callable) -> Callable:
    __f = click.option('-m', '--message', help='Comment text.')(__f)
    return __f


def attrib_parser(__f: Callable) -> Callable:
    __f = click.option('-o', '--order', default='number',
                       type=click.Choice(['number', 'updated']),
                       help='Sort order for listing bugs.')(__f)
    __f = click.option('-s', '--state', default='open',
                       type=click.Choice(['open', 'closed', 'all']),
                       help='State of bugs to operate on.')(__f)
    return __f


def stdin_parser(__f: Callable) -> Callable:
    __f = click.option('--stdin', is_flag=True,
                       help='Read message from standard input.')(__f)
    return __f


def text_parser(__f: Callable) -> Callable:
    __f = click.option('--title')(__f)
    __f = click.option('--body')(__f)
    return __f


def label_parser(__f: Callable) -> Callable:
    __f = click.option('-a', '--add', multiple=True,
                       help='Add label to issue.')(__f)
    __f = click.option('-c', '--create', multiple=True,
                       help='Create new label and add to issue.')(__f)
    return __f


@cli.command()
@click.option('--local/--no-local',
              help='Set access token for local repository only.')
@click.pass_obj
def setup(globs: AttrDict, local: bool):
    """Setup GitHub access token."""
    if not utils.SYSTEM_CERTS:
        warn('Falling back on bundled certificates')
    if utils.CURL_CERTS:
        warn('Using certs specified in $CURL_CERTS')
    default_user = os.getenv('GITHUB_USER',
                             utils.get_git_config_val('github.user',
                                                      getpass.getuser()))
    user = click.prompt('GitHub user', default_user)
    if not user:
        user = default_user
    password = click.prompt('GitHub password', hide_input=True,
                            confirmation_prompt=False)

    private = click.confirm('Support private repositories')

    data = {
        'scopes': ['repo' if private else 'public_repo'],
        'note': 'hubugs',
        'note_url': 'https://github.com/JNRowe/hubugs'
    }

    # Unfortunately, we need to forcibly define the header to workaround
    # GitHub sending a 404(in an effort to stop information leakage)
    header = {
        'Authorization': 'Basic ' + b64encode(':'.join([user, password]))
    }
    r, auth = globs.req_post('https://api.github.com/authorizations',
                             body=data, headers=header, model='Authorisation',
                             token=False)
    utils.set_git_config_val('hubugs.token', auth.token, local)
    success('Configuration complete!')


@cli.command(name='list')
@click.option('-l', '--label', multiple=True,
              help='List bugs with specified label.')
@click.option('-p', '--page', help='Page number.', type=click.INT,
              default=1)
@click.option('-r', '--pull-requests', is_flag=True,
              help='List only pull requests.')
@attrib_parser
@click.pass_obj
def list_bugs(globs: AttrDict, label: List[str], page: int,
              pull_requests: bool, order: str, state: str):
    """Listing bugs."""
    bugs = []
    params = {}
    if pull_requests:
        # FIXME: Dirty solution to supporting PRs only, needs rethink
        url = '{}/repos/{}/pulls'.format(globs.host_url, globs.project)
    else:
        url = ''
    if page != 1:
        params['page'] = page
    if label:
        params['labels'] = ','.join(label)

    states = ['open', 'closed'] if state == 'all' else [state, ]
    for state in states:
        _params = params.copy()
        _params['state'] = state
        r, _bugs = globs.req_get(url, params=_params, model='Issue')
        bugs.extend(_bugs)

    result = template.display_bugs(bugs, order, state=state,
                                   project=globs.repo_obj())
    if result:
        utils.pager(result, pager=globs.pager)


@cli.command()
@attrib_parser
@click.argument('term')
@click.pass_obj
def search(globs: AttrDict, order: str, state: str, term: str):
    """Searching bugs."""
    search_url = '{}/search/issues'.format(globs.host_url)
    states = ['open', 'closed'] if state == 'all' else [state, ]
    params = {
        'q': term,
        'repo': globs.project,
    }
    bugs = []
    for state in states:
        params['state'] = state
        r, c = globs.req_get(search_url, params=params, model='issue')
        bugs.extend(c.issues)
    result = template.display_bugs(bugs, order, term=term, state=state,
                                   project=globs.repo_obj())
    if result:
        utils.pager(result, pager=globs.pager)


@cli.command()
@click.option('-f', '--full/--no-full', help='Show bug including comments.')
@click.option('-p', '--patch/--no-patch', is_flag=True,
              help='Display patches for pull requests.')
@click.option('-o', '--patch-only', is_flag=True,
              help='Display only the patch content of pull requests.')
@click.option('-b', '--browse', is_flag=True,
              help='Open bug in web browser.')
@bugs_parser
@click.pass_obj
def show(globs: AttrDict, full: bool, patch: bool, patch_only: bool,
         browse: bool, bugs: List[int]):
    """Displaying bugs."""
    results = []
    tmpl = template.get_template('view', '/issue.txt')
    for bug_no in bugs:
        if browse:
            click.launch('https://github.com/{}/issues/{:d}'.format(
                globs.project, bug_no))
            continue
        r, bug = globs.req_get(bug_no, model='Issue')

        if full and bug.comments:
            r, comments = globs.req_get('{}/comments'.format(bug_no),
                                        model='Comment')
        else:
            comments = []
        if (patch or patch_only) and bug.pull_request:
            url = '{}/repos/{}/pulls/{}'.format(globs.host_url, globs.project,
                                                bug_no)
            headers = {'Accept': 'application/vnd.github.patch'}
            r, c = globs.req_get(url, headers=headers, is_json=False)
            patch = c.decode('utf-8')
        else:
            patch = None
        results.append(tmpl.render(bug=bug, comments=comments, full=True,
                                   patch=patch, patch_only=patch_only,
                                   project=globs.repo_obj()))
    if results:
        utils.pager('\n'.join(results), pager=globs.pager)


@cli.command(name='open')
@label_parser
@stdin_parser
@text_parser
@click.pass_obj
def open_bug(globs: AttrDict, add: List[str], create: List[str], stdin: bool,
             title: str, body: str):
    """Opening new bugs."""
    utils.sync_labels(globs, add, create)
    if stdin:
        text = click.get_text_stream('stdin').readlines()
    elif not title:
        text = template.edit_text('open').splitlines()
    if stdin or not title:
        title = text[0]
        body = '\n'.join(text[1:])
    else:
        title = title
        body = body
    data = {'title': title, 'body': body, 'labels': add + create}
    r, bug = globs.req_post('', body=data, model='Issue')
    success('Bug {:d} opened'.format(bug.number))


@cli.command()
@message_parser
@stdin_parser
@bugs_parser
@click.pass_obj
def comment(globs: AttrDict, message: str, stdin: bool, bugs: List[int]):
    """Commenting on bugs."""
    if stdin:
        message = click.get_text_stream().read()
    elif message:
        message = message
    else:
        message = template.edit_text()
    for bug in bugs:
        globs.req_post('{}/comments'.format(bug), body={'body': message},
                       model='Comment')


@cli.command()
@stdin_parser
@text_parser
@bugs_parser
@click.pass_obj
def edit(globs: AttrDict, stdin: bool, title: str, body:str, bugs: List[int]):
    """Editing bugs."""
    if (title or stdin) and len(bugs) > 1:
        raise ValueError('Can not use --stdin or command line title/body '
                         'with multiple bugs')
    for bug in bugs:
        if stdin:
            text = click.get_text_stream().readlines()
        elif not title:
            r, current = globs.req_get(bug, model='Issue')
            current_data = {'title': current.title, 'body': current.body}
            text = template.edit_text('open', current_data).splitlines()
        if stdin or not title:
            title = text[0]
            body = '\n'.join(text[1:])
        else:
            title = title
            body = body

        data = {'title': title, 'body': body}
        globs.req_post(bug, body=data, model='Issue')


@cli.command()
@stdin_parser
@message_parser
@bugs_parser
@click.pass_obj
def close(globs: AttrDict, stdin: bool, message: str, bugs: List[int]):
    """Closing bugs."""
    if stdin:
        message = click.get_text_stream().read()
    elif not message:
        try:
            message = template.edit_text()
        except template.EmptyMessageError:
            # Message isn't required for closing, but it is good practice
            message = None
    else:
        message = message
    for bug in bugs:
        if message:
            globs.req_post('{}/comments'.format(bug), body={'body': message},
                           model='Comment')
        globs.req_post(bug, body={'state': 'closed'}, model='Issue')


@cli.command()
@stdin_parser
@message_parser
@bugs_parser
@click.pass_obj
def reopen(globs: AttrDict, stdin: bool, message: str, bugs: List[int]):
    """Reopening closed bugs."""
    if stdin:
        message = click.get_text_stream().read()
    elif not message:
        try:
            message = template.edit_text()
        except template.EmptyMessageError:
            # Message isn't required for reopening, but it is good practice
            message = None
    for bug in bugs:
        if message:
            globs.req_post('{}/comments'.format(bug), body={'body': message},
                           model='Comment')
        globs.req_post(bug, body={'state': 'open'}, model='Issue')


@cli.command()
@label_parser
@click.option('-r', '--remove', multiple=True,
              help='Remove label from issue.')
@click.option('-l', '--list', is_flag=True, help='List available labels.')
@click.argument('bugs', nargs=-1, required=False, type=click.INT)
@click.pass_obj
def label(globs: AttrDict, add: List[str], create: List[str],
          remove: List[str], list: bool, bugs: List[int]):
    """Labelling bugs."""
    label_names = utils.sync_labels(globs, add, create)

    if list:
        click.echo(', '.join(sorted(label_names)))
        return

    for bug_no in bugs:
        r, bug = globs.req_get(bug_no, model='Issue')
        labels = [label.name for label in bug.labels]
        labels.extend(add + create)

        for string in remove:
            labels.remove(string)
        globs.req_post(bug_no, body={'labels': labels}, model='Label')


@cli.command()
@click.argument('milestone')
@click.argument('bugs', nargs=-1, required=False, type=click.INT)
@click.pass_obj
def milestone(globs: AttrDict, milestone: str, bugs: List[int]):
    """Issue milestones."""
    milestones_url = '{}/repos/{}/milestones'.format(globs.host_url,
                                                     globs.project)
    r, milestones = globs.req_get(milestones_url, model='Milestone')

    milestone_mapping = dict((m.title, m.number) for m in milestones)

    try:
        milestone = milestone_mapping[milestone]
    except KeyError:
        raise ValueError('No such milestone {:!r}'.format(milestone))

    for bug_no in bugs:
        globs.req_post(bug_no, body={'milestone': milestone},
                       model='Milestone')


@cli.command()
@click.option('-o', '--order', default='number',
              type=click.Choice(['number', 'due_date', 'completeness']),
              help='Sort order for listing bugs.')
@click.option('-s', '--state', default='open',
              type=click.Choice(['open', 'closed']),
              help='State of milestones to operate on.')
@click.option('-c', '--create', help='Create new milestone.')
@click.option('-l', '--list', is_flag=True,
              help='List available milestones.')
@click.pass_obj
def milestones(globs: AttrDict, order: str, state: str, create: str,
               list: bool):
    """Repository milestones."""
    if not list and not create:
        fail('No action specified!')
        return 1
    milestones_url = '{}/repos/{}/milestones'.format(globs.host_url,
                                                     globs.project)
    r, milestones = globs.req_get(milestones_url, model='Milestone')

    if list:
        tmpl = template.get_template('view', '/list_milestones.txt')
        columns = utils.T.width if utils.T.width else 80
        max_id = max(i.number for i in milestones)
        id_len = len(str(max_id))

        result = tmpl.render(milestones=milestones, order=order, state=state,
                             project=globs.repo_obj(),
                             id_len=id_len, max_title=columns - id_len - 2)
        if result:
            utils.pager(result, pager=globs.pager)
    elif create:
        data = {'title': create}
        r, milestone = globs.req_post('', body=data, model='Milestone')
        success('Milestone {:d} created'.format(milestone.number))


@cli.command()
@click.pass_obj
def report_bug(globs: AttrDict):
    """Report a new bug against hubugs."""
    local = globs.project == 'JNRowe/hubugs'
    globs.project = 'JNRowe/hubugs'

    import html2text, jinja2, pygments  # NOQA: E401
    versions = dict([(m.__name__, getattr(m, '__version__', 'No version info'))
                     for m in (click, html2text, httplib2, jinja2, pygments)])
    data = {
        'local': local,
        'sys': sys,
        'version': _version.dotted,
        'versions': versions,
        'certs': utils.CA_CERTS,
    }
    text = template.edit_text('hubugs_report', data).splitlines()
    title = text[0]
    body = '\n'.join(text[1:])

    data = {'title': title, 'body': body}
    r, bug = globs.req_post('', body=data, model='Issue')
    success('Bug {:d} opened against hubugs, thanks!'.format(bug.number))


def main() -> int:
    """Main command-line entry point.

    Returns:
        Exit code
    """
    try:
        cli()
    except utils.HttpClientError as error:
        fail(error.content[0])
        return errno.EINVAL
    except httplib2.ServerNotFoundError:
        fail('Project lookup failed.  Network or GitHub down?')
        return errno.ENXIO
    except (utils.RepoError) as error:
        fail(error.args[0])
        return errno.EINVAL
    except (EnvironmentError, ValueError) as error:
        fail(error.args[1])
        return errno.EINVAL



#
"""test_template - Test templating handling."""
# Copyright © 2011-2018  James Rowe <jnrowe@gmail.com>
#
# SPDX-License-Identifier: GPL-3.0+
#
# This file is part of hubugs.
#
# hubugs is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# hubugs is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# hubugs.  If not, see <http://www.gnu.org/licenses/>.

from collections import namedtuple
from datetime import (datetime, timedelta)
from typing import Dict, Optional

from html2text import __version__ as h2t_version
from mock import patch
from pygments import (formatters, lexers)
from pytest import mark, raises

from hubugs import template


@mark.parametrize('fg, bg, attributes, expected', [
    ('red', None, {}, '\x1b[31'),
    (None, 'blue', {}, '\x1b[44m'),
    (None, None, {'bold': True}, '\x1b[1m'),
])
def test_Colourise_color(fg: str, bg: Optional[str],
                         attributes: Dict[str, bool], expected: str):
    output = template.colourise('s', fg, bg, **attributes)
    assert expected in output


def test_Colourise_invalid_colour():
    with raises(TypeError):
        template.colourise('s', 'mauve with a hint of green')


def pyg_side_effect(*args, **kwargs):
    return namedtuple('Call', 'args kwargs')(args, kwargs)


def test_highlight(monkeypatch):
    monkeypatch.setattr('hubugs.template.pyg_highlight', pyg_side_effect)
    monkeypatch.setattr('sys.stdout.isatty', lambda: True)

    result = template.highlight('+++ a\n--- b\n+Test\n')
    assert isinstance(result.args[1], lexers.DiffLexer)
    assert isinstance(result.args[2], formatters.terminal.TerminalFormatter)


def test_highlight_lexer(monkeypatch):
    monkeypatch.setattr('hubugs.template.pyg_highlight', pyg_side_effect)
    monkeypatch.setattr('sys.stdout.isatty', lambda: True)

    result = template.highlight('True', 'python')
    assert isinstance(result.args[1], lexers.PythonLexer)


def test_highlight_formatter(monkeypatch):
    monkeypatch.setattr('hubugs.template.pyg_highlight', pyg_side_effect)
    monkeypatch.setattr('sys.stdout.isatty', lambda: True)

    result = template.highlight('True', formatter='terminal256')
    assert isinstance(result.args[2],
                      formatters.terminal256.Terminal256Formatter)


def test_EditText_no_message(monkeypatch):
    monkeypatch.setattr('click.edit', lambda *args, **kwargs: None)

    with raises(template.EmptyMessageError):
        template.edit_text()


def test_EditText_message(monkeypatch):
    monkeypatch.setattr('click.edit', lambda *args, **kwargs: 'Some message')

    assert template.edit_text() == 'Some message'


def test_EditText_message_prefill(monkeypatch):
    monkeypatch.setattr('click.edit', lambda t, *args, **kwargs: t)

    data = {'title': 'Some message'}
    assert template.edit_text('open', data) == data['title']


def test_Markdown_basic():
    assert template.markdown('### hello') == '<h3>hello</h3>\n'


def test_Html2Text_basic():
    assert template.html2text('<h3>hello</h3>') == '### hello'


def test_Html2Text_width():
    para = """<p>This is a long paragraph that needs wrapping to work so it
    doesn’t make you want to claw your eyes out."""
    assert template.html2text(para).count('\n') == 1
    # FIXME: Recent html2text version have changed API
    if isinstance(h2t_version, str) and h2t_version <= '2014.4.5':
        assert template.html2text(para, width=20).count('\n') == 1


@mark.parametrize('group, name', [
    ('edit', 'default.mkd'),
    ('view', 'issue.txt'),
])
def test_get_template(group: str, name: str):
    t = template.get_template(group, name)
    assert t.filename.endswith('/templates/default/%s/%s' % (group, name))


def test_jinja_filter(monkeypatch):
    monkeypatch.setattr('hubugs.template.ENV.filters', {})

    def null_func():
        pass

    template.jinja_filter(null_func)
    assert template.ENV.filters['null_func'] == null_func



#
"""conf - Sphinx configuration information."""
# Copyright © 2011-2018  James Rowe <jnrowe@gmail.com>
#
# SPDX-License-Identifier: GPL-3.0+
#
# This file is part of hubugs.
#
# hubugs is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# hubugs is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# hubugs.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
from contextlib import suppress
from subprocess import CalledProcessError, PIPE, run

root_dir = os.path.dirname(os.path.dirname(__file__))
sys.path.insert(0, root_dir)

import hubugs  # NOQA: E402

on_rtd = 'READTHEDOCS' in os.environ
if not on_rtd:
    import sphinx_rtd_theme

extensions = \
    [f'sphinx.ext.{ext}' for ext in ['autodoc', 'extlinks', 'intersphinx',
                                     'napoleon', 'todo']] \
    + ['sphinx_autodoc_typehints', ]

master_doc = 'index'
source_suffix = '.rst'

project = 'hubugs'
copyright = '2010-2016  James Rowe'

release = hubugs._version.dotted
version = release.rsplit('.', 1)[0]

html_experimental_html5_writer = True
modindex_common_prefix = ['hubugs.', ]

# readthedocs.org handles this setup for their builds, but it is nice to see
# approximately correct builds on the local system too
if not on_rtd:
    html_theme = "sphinx_rtd_theme"
    html_theme_path = [sphinx_rtd_theme.get_html_theme_path(), ]

pygments_style = 'sphinx'
with suppress(CalledProcessError):
    proc = run(['git', 'log', "--pretty=format:'%ad [%h]'", '--date=short',
                '-n1'],
               stdout=PIPE)
    html_last_updated_fmt = proc.stdout.decode()

man_pages = [
    ('hubugs.1', 'hubugs', 'hubugs Documentation', ['James Rowe', ], 1)
]

# Autodoc extension settings
autoclass_content = 'init'
autodoc_default_flags = ['members', ]

# intersphinx extension settings
intersphinx_mapping = {
    k: (v, os.getenv(f'SPHINX_{k.upper()}_OBJECTS'))
    for k, v in {
        'click': 'https://click.palletsprojects.com/en/6.x/',
        'jinja': 'http://jinja.pocoo.org/docs/',
        'jnrbase': 'https://jnrbase.readthedocs.io/en/latest/',
        'python': 'https://docs.python.org/3/',
}.items()}

extlinks = {
    'pypi': ('https://pypi.org/project/%s/', ''),
}

# todo extension settings
todo_include_todos = True

# napoleon extension settings
napoleon_numpy_docstring = False



#
"""utils - Utility functions for hubugs."""
# Copyright © 2011-2018  James Rowe <jnrowe@gmail.com>
#                        Matt Leighton <mleighy@gmail.com>
#
# SPDX-License-Identifier: GPL-3.0+
#
# This file is part of hubugs.
#
# hubugs is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# hubugs is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# hubugs.  If not, see <http://www.gnu.org/licenses/>.

import configparser
import json
import os
import re
import subprocess
import sys

from functools import partial
from typing import List, Optional
from urllib.parse import urlencode

import click
import httplib2

from jnrbase.attrdict import AttrDict
from jnrbase.colourise import warn
from jnrbase.xdg_basedir import user_cache

from . import (_version, models)

try:
    import ca_certs_locater
    CA_CERTS = ca_certs_locater.get()
except ImportError:
    CA_CERTS = None


class HttpClientError(ValueError):

    """Error raised for client error status codes."""

    def __init__(self, message, response, content):
        super(HttpClientError, self).__init__(message)
        self.response = response
        self.content = content


class RepoError(ValueError):

    """Error raised for invalid repository values."""


def get_github_api():
    """Create a GitHub API instance.

    Returns:
        httplib2.Http: GitHub HTTP session
    """
    cache_dir = user_cache('hubugs')
    with open('{}/CACHEDIR.TAG'.format(cache_dir), 'w') as f:
        f.writelines([
            'Signature: 8a477f597d28d172789f06886806bc55\n',
            '# This file is a cache directory tag created by hubugs.\n',
            '# For information about cache directory tags, see:\n',
            '#   http://www.brynosaurus.com/cachedir/\n',
            ])
    return httplib2.Http(cache_dir, ca_certs=CA_CERTS)


def get_git_config_val(__key: str, default: Optional[str] = None,
                       local_only: Optional[bool] = False) -> str:
    """Fetch a git configuration value.

    Args:
        __key: Configuration value to fetch
        default: Default value to use, if key isn’t set
        local_only: Fetch configuration values from repo config only

    Return:
        Git config value, if set
    """
    cmd = ['git', 'config', ]
    if local_only:
        cmd.append('--local')
    cmd.extend(['--get', __key])
    try:
        output = subprocess.check_output(cmd).decode().strip()
    except subprocess.CalledProcessError:
        output = default
    if output and output.startswith('!'):
        try:
            output = subprocess.check_output(output[1:].split())
            output = output.decode().strip()
        except subprocess.CalledProcessError:
            print('Whoops!')
            sys.exit(97)
    return output


def set_git_config_val(__key: str, __value: str,
                       local_only: Optional[bool] = False):
    """Set a git configuration value.

    Args:
        __key: Configuration value to fetch
        __value: Value to set
        local_only: Set configuration values from repo config only
    """
    cmd = ['git', 'config', ]
    if not local_only:
        cmd.append('--global')
    cmd.extend([__key, value])
    subprocess.check_output(cmd, stderr=subprocess.STDOUT)


def get_editor() -> List[str]:
    """Choose a suitable editor.

    See :manpage:`git-var(1)` for details.

    Return:
        Users chosen editor, or ``vi`` if not set
    """
    output = subprocess.check_output(['git', 'var', 'GIT_EDITOR'])
    return output.decode().strip().split()


def get_repo() -> str:
    """Extract GitHub project name from git/hg config.

    We check the git config for ``hubugs.project``, and then fall back to
    ``remote.origin.url``.  If both of these fail we check a mercurial root, to
    satisfy the ``hg-git`` users.

    Returns:
        GitHub project name, including user
    """
    data = get_git_config_val('hubugs.project', local_only=True)
    if data:
        return data
    data = get_git_config_val('remote.origin.url', local_only=True)
    if not data:
        try:
            root = subprocess.check_output(['hg', 'root'],
                                           stderr=subprocess.PIPE)
            root = root.decode().strip()
        except (OSError, subprocess.CalledProcessError):
            # No mercurial install, or not in a mercurial tree
            pass
        else:
            conf = configparser.ConfigParser()
            conf.read(os.path.join(root, '.hg', 'hgrc'))
            with contextlib.suppress(configparser.NoSectionError,
                                     configparser.NoOptionError):
                data = conf.get('paths', 'default')

    if not data:
        raise RepoError('Unable to guess project from repository')

    match = re.match(
        r"""
        (?:git(?:@|://)  # SSH or git protocol
          |git\+ssh://(?:git@)  # hg-git SSH URLs
          |https?://
           (?:.*@)?)  # HTTP URLs support optional auth data
        github.com[:/]  # hostname, : sep for SSH URL
        ([^/]+/.*?)  # project
        (?:.git)?$  # .git suffix is optional in all clone URLs
        """,
        data, re.VERBOSE)
    if match:
        return match.groups()[0]
    else:
        raise RepoError('Invalid project configuration, specify with '
                        "‘--project’ option")


def pager(__text: str, pager: Optional[bool] = False):
    """Pass output through pager.

    Args:
        __text: Text to page
        pager: Pager to use
    """
    if pager:
        click.echo_via_pager(__text)
    else:
        click.echo(__text)


def setup_environment(__project, __host_url):
    """Configure execution environment for commands dispatch."""
    env = AttrDict()

    if not __project:
        __project = get_repo()

    http = get_github_api()

    base_headers = {
        'Accept': 'application/vnd.github.v3+json',
        'User-Agent': _version.web,
    }

    token = os.getenv('HUBUGS_TOKEN', get_git_config_val('hubugs.token', None))
    if token:
        base_headers['Authorization'] = 'token {}'.format(token)

    def http_method(__url, method='GET', params=None, body=None, headers=None,
                    model=None, is_json=True, token=True):
        lheaders = base_headers.copy()
        if token and 'Authorization' not in lheaders:
            raise EnvironmentError('No hubugs authorisation token found!  '
                                   "Run ‘hubugs setup’ to create a token")
        if headers:
            lheaders.update(headers)
        if not isinstance(__url, str) or not __url.startswith('http'):
            __url = '{}/repos/{}/issues{}{}'.format(__host_url, __project,
                                                    '/' if __url else '',
                                                    __url)
        if params:
            __url += '?' + urlencode(params)
        if is_json and body:
            body = json.dumps(body)
        r, c = http.request(__url, method=method, body=body, headers=lheaders)
        if is_json:
            c = json.loads(c.decode('utf-8'),
                           object_hook=partial(models.object_hook,
                                               __name=model))
        if str(r.status)[0] == '4':
            raise HttpClientError(str(r.status), r, c)
        return r, c

    env['req_get'] = http_method
    env['req_post'] = partial(http_method, method='POST')

    def repo_obj():
        r, c = http_method('{}/repos/{}'.format(__host_url, __project),
                           model='Repo')
        if not c.has_issues:
            raise RepoError(
                "Issues aren’t enabled for {:!r}".format(__project))
    env['repo_obj'] = repo_obj
    return env


def sync_labels(__globs: AttrDict, __add, __create) -> List[str]:
    """Manage labels for a project.

    Args:
        globs: Global argument configuration

    Returns:
        List of project’s label names
    """
    labels_url = '{}/repos/{}/labels'.format(__globs.host_url, __globs.project)
    r, c = globs.req_get(labels_url, model='Label')
    label_names = [label.name for label in c]

    for label in __add:
        if label not in label_names:
            raise ValueError('No such label {!r}'.format(label))
    for label in __create:
        if label in label_names:
            warn('{!r} label already exists'.format(label))
        else:
            data = {'name': label, 'color': '000000'}
            globs.req_post(labels_url, body=data, model='Label')
    return label_names + add + create
